/*
** Module   : TemporaryAccountPool.java
** Abstract : temporary account pool implementation.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20131212 Created initial version
*/

package com.goldencode.p2j.main;

import java.util.concurrent.*;
import java.util.logging.*;

import com.goldencode.p2j.admin.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.util.*;

/**
 * A temporary account pool implementation.
 * On server start-up all temporary accounts which belongs to "temps" group are deleted.
 * New accounts are created and added to pool. The new accounts are created disabled.
 * Because the password is stored in directory as a hash string, the clear text password 
 * is kept in memory (pool) inside a <code>TemporaryAccount</code> structure.
 * When an item is requested from pool the item is removed form pool and the associated
 * account is enabled before returned to caller.
 * If the pool is empty the current thread will wait for an specified amount of time to
 * get an item returned to pool by other threads.
 * When an item is returned to pool first the associated account is disabled than the item
 * is enqueued at the tail of the pool and other threads waiting for an item are notified.
 * In order to manipulate user accounts special permissions are mandatory.
 * A dedicated worker thread is spawned on start-up and is used to manage user accounts at 
 * runtime. This daemon thread is running on server security context having admin permissions.
 *  
 * @author mag
 *
 */
public class TemporaryAccountPool
{   
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(TemporaryAccountWorker.class.getName());
      
   /** Temporary accounts group name */
   private static final String TEMP_GROUP = "temps";
   
   /** Random generator length */
   public static final int GEN_LENGTH = 16; 

   /** Random generator digits */
   public static final int GEN_DIGITS = 8; 
   
   /** Random generator symbols */   
   public static final int GEN_SYMBOLS = 4; 
   
   /** FIFO Pool */
   private static ConcurrentLinkedQueue<TemporaryAccount> pool = 
            new ConcurrentLinkedQueue<>();
   
   /** Pool Manager instance */
   private static volatile TemporaryAccountPool poolManager;
      
   /** Task used to change the user account status */
   private static TemporaryAccountWorker poolTask;

   /** A daemon worker thread having a security context which allow to change user account. */
   private static AssociatedThread worker;
   
   /**
    * Initialize pool. This happens on server start-up.
    * All temporary accounts are deleted a new accounts are created.
    */
   static
   {      
      AdminServerImpl.setTargetLive(true);
      // Create TEMP_GROUP group if not exists
      createGroupIfNotExists();
      // Delete temporary accounts
      deleteTemporaryAccounts();
      // Create new temporary accounts
      createTemporaryAccounts();
      // Refresh target
      AdminServerImpl.targetRefresh();
      // Start worker thread
      startUp();
      // Test if pool has been created
      if (pool.isEmpty())
      {
         String message = "Admin privileges are mandatory to create temporary account pool.";
         LOG.logp(Level.WARNING,
                  "TemporaryAccountPool",
                  "static initializer",
                  LogHelper.generate(message));
      }
   }
   
   /**
    * Create new temporary accounts
    */
   private static void createTemporaryAccounts()
   {
      int poolSize = ClientBuilderOptions.getInstance().getPoolSize(); 
      for (int i = 0;i < poolSize;i++)
      {
         TemporaryAccount account = createUser(false);
         if (account != null)
         {
            pool.add(account);
         }
      }      
   }
   
   /**
    * Delete all temporary accounts. 
    * Temporary accounts belong to TEMP_GROUP group.
    */
   private static void deleteTemporaryAccounts()
   {
      TaggedName[] users = AdminServerImpl.listGroupUsers(TEMP_GROUP);
      if (users != null)
      {
         for (TaggedName name : users)
         {
            UserDef user = AdminServerImpl.getUser(name.getName());
            if (user != null)
            {
               AdminServerImpl.deleteUser(user.subjectId);
            }
         }
      }
   }
   
   /**
    * Check if temporary group exists. If not create group.
    */
   private static void createGroupIfNotExists()
   {
      GroupDef groupDef = AdminServerImpl.getGroup(TEMP_GROUP);
      if (groupDef == null)
      {
         groupDef = new GroupDef();
         groupDef.subjectId = TEMP_GROUP;
         groupDef.description = "all temps";  
         AdminServerImpl.addGroup(groupDef);
      }
   }
     
   /**
    * Create a temporary user.
    * 
    * @param   enabled
    *          Initial account status.
    *          
    * @return  Created user account.
    */
   static TemporaryAccount createUser(boolean enabled)
   {
      TemporaryAccount account = null;
      // Generate random values for account
      String subject = RandomWordGenerator.create(GEN_LENGTH, GEN_DIGITS, 0);
      String password = RandomWordGenerator.create(GEN_LENGTH, GEN_DIGITS, GEN_SYMBOLS);
      // Fill account structure
      UserDef user = new UserDef();
      user.enabled = enabled;
      user.protect = true;
      user.person = "chui";
      user.subjectId = subject;
      user.password = HashPassword.hashPassword(password);
      user.mode = 1;
      user.groups = new String[] { TEMP_GROUP };
      // Create temporary account
      if (AdminServerImpl.addUser(user, null))
      {
         account = new TemporaryAccount(user.subjectId, password);
      }
      // Return created account.
      return account;
   }

   /**
    * Create and return the pool manager instance.
    * This is a singleton design pattern. 
    * 
    * @return  An instance of TemporaryAccountPool
    */
   public static TemporaryAccountPool getInstance()
   {
      synchronized (TemporaryAccountPool.pool)
      {
         if (poolManager == null)
         {
            poolManager = new TemporaryAccountPool();
         }
      }
      return poolManager;
   }
   
   /**
    * Startup worker thread.
    */
   private static void startUp()
   {
      poolTask = new TemporaryAccountWorker();
      worker = new AssociatedThread(poolTask);
      worker.setDaemon(true);
      worker.start();      
   }

   /**
    * Shut down worker thread.
    */
   public static void shutDown()
   {
      try
      {
         poolTask.terminate();
         worker.interrupt();
         worker.join();
      }
      catch (InterruptedException e)
      {
         LOG.logp(Level.WARNING,
                  "TemporaryAccountPool.shutDowns()",
                  "",
                  LogHelper.generate("InterruptedException!"),
                  e);         
      }
   }
   
   /**
    * Get and remove an item from pool.
    * Enable the temporary account before return to caller. 
    * The item should be returned back to pool when no longer used.
    * On empty pool wait for an amount of time to get an item from pool.
    * If no items are available after waiting a null value is returned.
    * 
    * @return  The item at the head of pool or null if pool is empty and
    *          no item is available after waiting an amount of time. 
    */
   public TemporaryAccount poll()
   {
      synchronized (TemporaryAccountPool.pool)
      {
         // Get an item from pool or null if pool is empty.
         TemporaryAccount item = TemporaryAccountPool.pool.poll();
         // If item is null means pool is empty. 
         if (item == null)
         {
            try
            {
               // Max time to wait for a pool item. 
               long timeout = ClientBuilderOptions.getInstance().getTimeout();
               // Wait until an item will be available but no more than an amount of time. 
               while (timeout > 0)
               {
                  // Wait for notifications
                  long start = System.currentTimeMillis();
                  TemporaryAccountPool.pool.wait(timeout);
                  timeout -= (System.currentTimeMillis() - start);
                  // Try to get the item from pool
                  item = TemporaryAccountPool.pool.poll();
                  if (item != null)
                  {
                     // Enable account
                     enableAccount(item, true);
                     break;
                  }
               }
            }
            catch (InterruptedException e)
            {
               LOG.logp(Level.WARNING,
                        "TemporaryAccountPool.poll()",
                        "",
                        LogHelper.generate("InterruptedException!"),
                        e);         
            }
         }         
         else
         {
            // Enable account
            enableAccount(item, true);
         }

         return item;
      }
   }

   /**
    * Disable temporary account and return the item to pool.
    * The item is enqueued at the tail of the pool. 
    * Notify other threads waiting for a pool item.
    * 
    * @param   item
    *          A TemporaryAccount item returned to pool.
    */
   public void tail(TemporaryAccount item)
   {
      synchronized (TemporaryAccountPool.pool)
      {
         // Change user account status
         enableAccount(item, false);
         TemporaryAccountPool.pool.add(item);
         TemporaryAccountPool.pool.notifyAll();
      }
   }
   
   /**
    * Enable / disable temporary account.
    * 
    * @param   item
    *          Account pool item.
    * @param   enabled
    *          <code>true</code> enable account.
    *          <code>false</code> disable account.
    */
   private void enableAccount(TemporaryAccount item, boolean enabled)
   {
      // Change user account status
      Runnable core = new UpdateAccountTask(item, enabled);
      TemporaryAccountPool.poolTask.execute(core);               
   }
}
